/** @module hareactive/stream */
"use strict";
const Behavior_1 = require("./Behavior");
/**
 * A stream is a list of occurences over time. Each occurence happens
 * at a discrete point in time and has an associated value.
 * Semantically it is a list `type Stream<A> = [Time, A]`.
 */
class Stream {
    constructor(data, push) {
        this.eventListeners = [];
        this.data = data;
        this.push = push;
    }
    publish(a) {
        const list = this.eventListeners;
        const l = this.eventListeners.length;
        for (let i = 0; i < l; ++i) {
            const e = list[i];
            e.push(a, e.data, e);
        }
    }
    ;
    subscribe(fn) {
        this.eventListeners.push(new Stream({}, fn));
    }
    map(fn) {
        const e = new MapStream(fn);
        this.eventListeners.push(e);
        return e;
    }
    // mapTo<B>(val: B): Stream<B> {
    //   const s = new MapToStream(val);
    //   this.eventListeners.push(s);
    //   return s;
    // }
    // merge<B>(otherStream: Stream<B>): Stream<(A|B)> {
    //   const e = new SinkStream<(A|B)>();
    //   this.eventListeners.push(e);
    //   otherStream.eventListeners.push(e);
    //   return e;
    // }
    filter(fn) {
        const e = new Stream(new FilterData(fn), pushFilter);
        this.eventListeners.push(e);
        return e;
    }
    scanS(fn, startingValue) {
        return scanS(fn, startingValue, this);
    }
    scan(fn, init) {
        return Behavior_1.scan(fn, init, this);
    }
    unlisten(listener) {
        const l = this.eventListeners;
        const idx = l.indexOf(listener);
        if (idx !== -1) {
            if (idx !== l.length - 1) {
                l[idx] = l[l.length - 1];
            }
            l.length--; // remove the last element of the list
        }
    }
}
exports.Stream = Stream;
/** @private */
function pushSink(a, d, s) {
    s.publish(a);
}
class MapData {
    constructor(fn) {
        this.fn = fn;
    }
}
function pushMap(a, d, s) {
    s.publish(d.fn(a));
}
// class MapToStream<A> extends Stream<A> {
//   constructor(private val: A) { super(); }
//   push(a: any): void {
//     this.publish(this.val);
//   }
// }
class FilterData {
    constructor(fn) {
        this.fn = fn;
    }
}
function pushFilter(a, d, s) {
    if (d.fn(a)) {
        s.publish(a);
    }
}
/**
 * @param fn A predicate function that returns a boolean for `A`.
 * @param stream The stream to filter.
 * @returns Stream that only contains the occurences from `stream`
 * for which `fn` returns true.
 */
function filter(fn, stream) {
    return stream.filter(fn);
}
exports.filter = filter;
class ScanData {
    constructor(fn, last) {
        this.fn = fn;
        this.last = last;
    }
}
function pushScan(a, d, s) {
    const val = d.last = d.fn(a, d.last);
    s.publish(val);
}
/**
 * The returned  initially has the initial value, on each
 * occurence in `source` the function is applied to the current value
 * of the behaviour and the value of the occurence, the returned value
 * becomes the next value of the behavior.
 */
function scanS(fn, startingValue, stream) {
    return Behavior_1.fromFunction(() => {
        const s = new Stream(new ScanData(fn, startingValue), pushScan);
        stream.eventListeners.push(s);
        return s;
    });
}
exports.scanS = scanS;
// class SnapshotStream<A, B> extends Stream<[A, B]> {
//   constructor(private behavior: Behavior<B>, stream: Stream<A>) {
//     super();
//     stream.eventListeners.push(this);
//   }
//   push(a: A): void {
//     this.publish([a, at(this.behavior)]);
//   }
// }
// export function snapshot<A, B>(behavior: Behavior<B>, stream: Stream<A>): Stream<[A, B]> {
//   return new SnapshotStream(behavior, stream);
// }
// class SnapshotWithStream<A, B, C> extends Stream<C> {
//   constructor(
//     private fn: (a: A, b: B) => C,
//     private behavior: Behavior<B>,
//     stream: Stream<A>
//   ) {
//     super();
//     stream.eventListeners.push(this);
//   }
//   push(a: A): void {
//     this.publish(this.fn(a, at(this.behavior)));
//   }
// }
// export function snapshotWith<A, B, C>(
//   fn: (a: A, b: B) => C,
//   behavior: Behavior<B>,
//   stream: Stream<A>
// ): Stream<C> {
//   return new SnapshotWithStream(fn, behavior, stream);
// }
// class SwitchBehaviorStream<A> extends Stream<A> {
//   private currentSource: Stream<A>;
//   constructor(private b: Behavior<Stream<A>>) {
//     super();
//     b.addListener(this);
//     const cur = this.currentSource = at(b);
//     cur.eventListeners.push(this);
//   }
//   push(a: any, changer: any): void {
//     if (changer === this.b) {
//       this.doSwitch(a);
//     } else {
//       this.publish(a);
//     }
//   }
//   private doSwitch(newStream: Stream<A>): void {
//     this.currentSource.unlisten(this);
//     newStream.eventListeners.push(this);
//     this.currentSource = newStream;
//   }
// }
/**
 * Takes a stream valued behavior and returns at stream that emits
 * values from the current stream at the behavior.
 */
// export function switchStream<A>(b: Behavior<Stream<A>>): Stream<A> {
//   return new SwitchBehaviorStream(b);
// }
// export function mergeList<A>(ss: Stream<A>[]): Stream<A> {
//   return ss.reduce((s1, s2) => s1.merge(s2), empty());
// }
/**
 * @returns A stream that never has any occurrences.
 */
function empty() {
    return new Stream({}, pushSink);
}
exports.empty = empty;
function subscribe(fn, stream) {
    stream.subscribe(fn);
}
exports.subscribe = subscribe;
function publish(a, stream) {
    stream.publish(a);
}
exports.publish = publish;
// export function merge<A, B>(a: Stream<A>, b: Stream<B>): Stream<(A|B)> {
//   return a.merge(b);
// }
function map(fn, stream) {
    return stream.map(fn);
}
exports.map = map;
function isStream(obj) {
    return (obj instanceof Stream);
}
exports.isStream = isStream;
//# sourceMappingURL=FailStream.js.map