/** @module hareactive/stream */
"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var frp_common_1 = require("./frp-common");
var Behavior_1 = require("./Behavior");
/**
 * A stream is a list of occurences over time. Each occurence happens
 * at a discrete point in time and has an associated value.
 * Semantically it is a list `type Stream<A> = [Time, A]`.
 */
var Stream = (function (_super) {
    __extends(Stream, _super);
    function Stream() {
        _super.call(this);
    }
    Stream.prototype.map = function (fn) {
        var s = new MapStream(fn);
        this.addListener(s);
        return s;
    };
    Stream.prototype.mapTo = function (val) {
        var s = new MapToStream(val);
        this.addListener(s);
        return s;
    };
    Stream.prototype.merge = function (otherStream) {
        var s = new SinkStream();
        this.addListener(s);
        otherStream.addListener(s);
        return s;
    };
    Stream.prototype.filter = function (fn) {
        var s = new FilterStream(fn);
        this.addListener(s);
        return s;
    };
    Stream.prototype.scanS = function (fn, startingValue) {
        var _this = this;
        return Behavior_1.fromFunction(function () { return new ScanStream(fn, startingValue, _this); });
    };
    Stream.prototype.scan = function (fn, init) {
        return Behavior_1.scan(fn, init, this);
    };
    return Stream;
}(frp_common_1.Reactive));
exports.Stream = Stream;
/** @private */
var SinkStream = (function (_super) {
    __extends(SinkStream, _super);
    function SinkStream() {
        _super.apply(this, arguments);
    }
    SinkStream.prototype.push = function (a) {
        this.child.push(a);
    };
    return SinkStream;
}(Stream));
exports.SinkStream = SinkStream;
var MapStream = (function (_super) {
    __extends(MapStream, _super);
    function MapStream(fn) {
        _super.call(this);
        this.fn = fn;
    }
    MapStream.prototype.push = function (a) {
        this.child.push(this.fn(a));
    };
    return MapStream;
}(Stream));
function map(fn, stream) {
    return stream.map(fn);
}
exports.map = map;
var MapToStream = (function (_super) {
    __extends(MapToStream, _super);
    function MapToStream(val) {
        _super.call(this);
        this.val = val;
    }
    MapToStream.prototype.push = function (a) {
        this.child.push(this.val);
    };
    return MapToStream;
}(Stream));
var FilterStream = (function (_super) {
    __extends(FilterStream, _super);
    function FilterStream(fn) {
        _super.call(this);
        this.fn = fn;
    }
    FilterStream.prototype.push = function (a) {
        if (this.fn(a) === true) {
            this.child.push(a);
        }
    };
    return FilterStream;
}(Stream));
/**
 * @param fn A predicate function that returns a boolean for `A`.
 * @param stream The stream to filter.
 * @returns Stream that only contains the occurences from `stream`
 * for which `fn` returns true.
 */
function filter(fn, stream) {
    return stream.filter(fn);
}
exports.filter = filter;
var ScanStream = (function (_super) {
    __extends(ScanStream, _super);
    function ScanStream(fn, last, source) {
        _super.call(this);
        this.fn = fn;
        this.last = last;
        source.addListener(this);
    }
    ScanStream.prototype.push = function (a) {
        var val = this.last = this.fn(a, this.last);
        this.child.push(val);
    };
    return ScanStream;
}(Stream));
/**
 * The returned  initially has the initial value, on each
 * occurence in `source` the function is applied to the current value
 * of the behaviour and the value of the occurence, the returned value
 * becomes the next value of the behavior.
 */
function scanS(fn, startingValue, stream) {
    return Behavior_1.fromFunction(function () { return new ScanStream(fn, startingValue, stream); });
}
exports.scanS = scanS;
var SnapshotStream = (function (_super) {
    __extends(SnapshotStream, _super);
    function SnapshotStream(behavior, stream) {
        _super.call(this);
        this.behavior = behavior;
        stream.addListener(this);
    }
    SnapshotStream.prototype.push = function (a) {
        this.child.push(Behavior_1.at(this.behavior));
    };
    return SnapshotStream;
}(Stream));
function snapshot(behavior, stream) {
    return new SnapshotStream(behavior, stream);
}
exports.snapshot = snapshot;
var SnapshotWithStream = (function (_super) {
    __extends(SnapshotWithStream, _super);
    function SnapshotWithStream(fn, behavior, stream) {
        _super.call(this);
        this.fn = fn;
        this.behavior = behavior;
        stream.child = this;
    }
    SnapshotWithStream.prototype.push = function (a) {
        this.child.push(this.fn(a, Behavior_1.at(this.behavior)));
    };
    return SnapshotWithStream;
}(Stream));
function snapshotWith(fn, behavior, stream) {
    return new SnapshotWithStream(fn, behavior, stream);
}
exports.snapshotWith = snapshotWith;
/** @private */
var SwitchOuter = (function () {
    function SwitchOuter(s) {
        this.s = s;
    }
    ;
    SwitchOuter.prototype.beginPulling = function () {
        throw new Error("not implemented");
    };
    SwitchOuter.prototype.endPulling = function () {
        throw new Error("not implemented");
    };
    SwitchOuter.prototype.push = function (a) {
        this.s.doSwitch(a);
    };
    return SwitchOuter;
}());
var SwitchBehaviorStream = (function (_super) {
    __extends(SwitchBehaviorStream, _super);
    function SwitchBehaviorStream(b) {
        _super.call(this);
        this.b = b;
        this.outerConsumer = new SwitchOuter(this);
        b.addListener(this.outerConsumer);
        var cur = this.currentSource = Behavior_1.at(b);
        cur.addListener(this);
    }
    SwitchBehaviorStream.prototype.push = function (a) {
        this.child.push(a);
    };
    SwitchBehaviorStream.prototype.doSwitch = function (newStream) {
        this.currentSource.removeListener(this);
        newStream.addListener(this);
        this.currentSource = newStream;
    };
    return SwitchBehaviorStream;
}(Stream));
/**
 * Takes a stream valued behavior and returns at stream that emits
 * values from the current stream at the behavior. I.e. the returned
 * stream always "switches" to the current stream at the behavior.
 */
function switchStream(b) {
    return new SwitchBehaviorStream(b);
}
exports.switchStream = switchStream;
var ChangesStream = (function (_super) {
    __extends(ChangesStream, _super);
    function ChangesStream(b) {
        _super.call(this);
        this.b = b;
        b.addListener(this);
    }
    ChangesStream.prototype.push = function (a) {
        this.child.push(a);
    };
    ChangesStream.prototype.beginPulling = function () {
        throw new Error("Cannot get changes from pulling behavior");
    };
    ChangesStream.prototype.endPulling = function () {
        throw new Error("Cannot get changes from pulling behavior");
    };
    return ChangesStream;
}(Stream));
function changes(b) {
    return new ChangesStream(b);
}
exports.changes = changes;
function mergeList(ss) {
    // FIXME: More performant implementation with benchmark
    return ss.reduce(function (s1, s2) { return s1.merge(s2); }, empty());
}
exports.mergeList = mergeList;
/**
 * @returns A stream that never has any occurrences.
 */
function empty() {
    return new SinkStream();
}
exports.empty = empty;
function subscribe(fn, stream) {
    stream.subscribe(fn);
}
exports.subscribe = subscribe;
function publish(a, stream) {
    stream.push(a);
}
exports.publish = publish;
function merge(a, b) {
    return a.merge(b);
}
exports.merge = merge;
function isStream(obj) {
    return (obj instanceof Stream);
}
exports.isStream = isStream;
//# sourceMappingURL=Stream.js.map