import { Monad } from "jabz/monad";
import { Observer } from "./frp-common";
import { Future } from "./future";
import { Stream } from "./stream";
/**
 * A behavior is a value that changes over time. Conceptually it can
 * be though of as a function from time to a value. I.e. `type
 * Behavior<A> = (t: Time) => A`.
 */
export declare abstract class Behavior<A> implements Observer<A>, Monad<A> {
    pushing: boolean;
    last: A;
    nrOfListeners: number;
    child: Observer<any>;
    abstract push(a: any): void;
    abstract pull(): A;
    constructor();
    map<B>(fn: (a: A) => B): Behavior<B>;
    mapTo<A>(v: A): Behavior<A>;
    static of<A>(v: A): Behavior<A>;
    of<A>(v: A): Behavior<A>;
    ap<B>(f: Behavior<(a: A) => B>): Behavior<B>;
    lift<T1, R>(f: (t: T1) => R, m: Behavior<T1>): Behavior<R>;
    lift<T1, T2, R>(f: (t: T1, u: T2) => R, m1: Behavior<T1>, m2: Behavior<T2>): Behavior<R>;
    lift<T1, T2, T3, R>(f: (t1: T1, t2: T2, t3: T3) => R, m1: Behavior<T1>, m2: Behavior<T2>, m3: Behavior<T3>): Behavior<R>;
    static multi: boolean;
    multi: boolean;
    chain<B>(fn: (a: A) => Behavior<B>): Behavior<B>;
    flatten: <B>() => Behavior<B>;
    endPulling(): void;
    beginPulling(): void;
    subscribe(cb: (a: A) => void): Observer<A>;
    addListener(c: Observer<A>): void;
    removeListener(listener: Observer<any>): void;
    observe(push: (a: A) => void, beginPulling: () => void, endPulling: () => void): CbObserver<A>;
    at(): A;
}
export declare function at<B>(b: Behavior<B>): B;
/**
 * Apply a function valued behavior to a value behavior.
 *
 * @param fnB behavior of functions from `A` to `B`
 * @param valB A behavior of `A`
 * @returns Behavior of the function in `fnB` applied to the value in `valB`
 */
export declare function ap<A, B>(fnB: Behavior<(a: A) => B>, valB: Behavior<A>): Behavior<B>;
/**
 * Creates a behavior for imperative impure pushing.
 */
export declare function sink<A>(initialValue: A): Behavior<A>;
/**
 * A placeholder behavior is a behavior without any value. It is used
 * to do value recursion in `./framework.ts`.
 * @private
 */
export declare class PlaceholderBehavior<B> extends Behavior<B> {
    private source;
    constructor();
    push(v: B): void;
    pull(): B;
    replaceWith(b: Behavior<B>): void;
}
export declare function placeholder(): PlaceholderBehavior<any>;
export declare function when(b: Behavior<boolean>): Behavior<Future<{}>>;
export declare function snapshotAt<A>(b: Behavior<A>, f: Future<any>): Behavior<Future<A>>;
export declare function switcher<A>(init: Behavior<A>, next: Future<Behavior<A>>): Behavior<A>;
export declare function stepper<B>(initial: B, steps: Stream<B>): Behavior<B>;
export declare function scan<A, B>(fn: (a: A, b: B) => B, init: B, source: Stream<A>): Behavior<Behavior<B>>;
export declare function toggle(initial: boolean, turnOn: Stream<any>, turnOff: Stream<any>): Behavior<boolean>;
export declare function fromFunction<B>(fn: () => B): Behavior<B>;
export declare class CbObserver<A> implements Observer<A> {
    private _push;
    private _beginPulling;
    private _endPulling;
    private source;
    constructor(_push: (a: A) => void, _beginPulling: () => void, _endPulling: () => void, source: Behavior<A>);
    push(a: A): void;
    beginPulling(): void;
    endPulling(): void;
}
/**
 * Observe a behavior for the purpose of executing imperative actions
 * based on the value of the behavior.
 */
export declare function observe<A>(push: (a: A) => void, beginPulling: () => void, endPulling: () => void, b: Behavior<A>): CbObserver<A>;
/**
 * Imperatively push a value into a behavior.
 */
export declare function publish<A>(a: A, b: Behavior<A>): void;
export declare function isBehavior(b: any): b is Behavior<any>;
export declare type Time = number;
export declare const time: Behavior<Time>;
export declare const timeFrom: Behavior<Behavior<Time>>;
