/** @module hareactive/stream */
"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var frp_common_1 = require("./frp-common");
var behavior_1 = require("./behavior");
/**
 * A stream is a list of occurences over time. Each occurence happens
 * at a discrete point in time and has an associated value.
 * Semantically it is a list `type Stream<A> = [Time, A]`.
 */
var Stream = (function (_super) {
    __extends(Stream, _super);
    function Stream() {
        _super.call(this);
    }
    Stream.prototype.map = function (fn) {
        var s = new MapStream(fn);
        this.addListener(s);
        return s;
    };
    Stream.prototype.mapTo = function (val) {
        var s = new MapToStream(val);
        this.addListener(s);
        return s;
    };
    Stream.prototype.combine = function (otherStream) {
        var s = new SinkStream();
        this.addListener(s);
        otherStream.addListener(s);
        return s;
    };
    Stream.prototype.filter = function (fn) {
        var s = new FilterStream(fn);
        this.addListener(s);
        return s;
    };
    Stream.prototype.scanS = function (fn, startingValue) {
        var _this = this;
        return behavior_1.fromFunction(function () { return new ScanStream(fn, startingValue, _this); });
    };
    Stream.prototype.scan = function (fn, init) {
        return behavior_1.scan(fn, init, this);
    };
    Stream.prototype.delay = function (ms) {
        var s = new DelayStream(ms);
        this.addListener(s);
        return s;
    };
    Stream.prototype.throttle = function (ms) {
        var s = new ThrottleStream(ms);
        this.addListener(s);
        return s;
    };
    Stream.prototype.debounce = function (ms) {
        var s = new DebounceStream(ms);
        this.addListener(s);
        return s;
    };
    return Stream;
}(frp_common_1.Reactive));
exports.Stream = Stream;
/** @private */
var SinkStream = (function (_super) {
    __extends(SinkStream, _super);
    function SinkStream() {
        _super.apply(this, arguments);
    }
    SinkStream.prototype.push = function (a) {
        this.child.push(a);
    };
    return SinkStream;
}(Stream));
exports.SinkStream = SinkStream;
var MapStream = (function (_super) {
    __extends(MapStream, _super);
    function MapStream(fn) {
        _super.call(this);
        this.fn = fn;
    }
    MapStream.prototype.push = function (a) {
        this.child.push(this.fn(a));
    };
    return MapStream;
}(Stream));
var MapToStream = (function (_super) {
    __extends(MapToStream, _super);
    function MapToStream(val) {
        _super.call(this);
        this.val = val;
    }
    MapToStream.prototype.push = function (a) {
        this.child.push(this.val);
    };
    return MapToStream;
}(Stream));
var FilterStream = (function (_super) {
    __extends(FilterStream, _super);
    function FilterStream(fn) {
        _super.call(this);
        this.fn = fn;
    }
    FilterStream.prototype.push = function (a) {
        if (this.fn(a) === true) {
            this.child.push(a);
        }
    };
    return FilterStream;
}(Stream));
var DelayStream = (function (_super) {
    __extends(DelayStream, _super);
    function DelayStream(ms) {
        _super.call(this);
        this.ms = ms;
    }
    DelayStream.prototype.push = function (a) {
        var _this = this;
        setTimeout(function () { return _this.child.push(a); }, this.ms);
    };
    return DelayStream;
}(Stream));
var ThrottleStream = (function (_super) {
    __extends(ThrottleStream, _super);
    function ThrottleStream(ms) {
        _super.call(this);
        this.ms = ms;
        this.isSilenced = false;
    }
    ThrottleStream.prototype.push = function (a) {
        var _this = this;
        if (!this.isSilenced) {
            this.child.push(a);
            this.isSilenced = true;
            setTimeout(function () {
                _this.isSilenced = false;
            }, this.ms);
        }
    };
    return ThrottleStream;
}(Stream));
var DebounceStream = (function (_super) {
    __extends(DebounceStream, _super);
    function DebounceStream(ms) {
        _super.call(this);
        this.ms = ms;
        this.timer = undefined;
    }
    DebounceStream.prototype.push = function (a) {
        var _this = this;
        clearTimeout(this.timer);
        this.timer = setTimeout(function () {
            _this.child.push(a);
        }, this.ms);
    };
    return DebounceStream;
}(Stream));
function debounce(ms, stream) {
    return stream.debounce(ms);
}
exports.debounce = debounce;
function throttle(ms, stream) {
    return stream.throttle(ms);
}
exports.throttle = throttle;
function delay(ms, stream) {
    return stream.delay(ms);
}
exports.delay = delay;
function apply(behavior, stream) {
    return stream.map(function (a) { return behavior_1.at(behavior)(a); });
}
exports.apply = apply;
/**
 * @param fn A predicate function that returns a boolean for `A`.
 * @param stream The stream to filter.
 * @returns Stream that only contains the occurences from `stream`
 * for which `fn` returns true.
 */
function filter(predicate, s) {
    return s.filter(predicate);
}
exports.filter = filter;
function split(predicate, stream) {
    // It should be possible to implement this in a faster way where
    // `predicate` is only called once for each occurrence
    return [stream.filter(predicate), stream.filter(function (a) { return !predicate(a); })];
}
exports.split = split;
function filterApply(predicate, stream) {
    return stream.filter(function (a) { return behavior_1.at(predicate)(a); });
}
exports.filterApply = filterApply;
function keepWhen(stream, behavior) {
    return stream.filter(function (_) { return behavior_1.at(behavior); });
}
exports.keepWhen = keepWhen;
var ScanStream = (function (_super) {
    __extends(ScanStream, _super);
    function ScanStream(fn, last, source) {
        _super.call(this);
        this.fn = fn;
        this.last = last;
        source.addListener(this);
    }
    ScanStream.prototype.push = function (a) {
        var val = this.last = this.fn(a, this.last);
        this.child.push(val);
    };
    return ScanStream;
}(Stream));
/**
 * The returned  initially has the initial value, on each
 * occurence in `source` the function is applied to the current value
 * of the behaviour and the value of the occurence, the returned value
 * becomes the next value of the behavior.
 */
function scanS(fn, startingValue, stream) {
    return stream.scanS(fn, startingValue);
}
exports.scanS = scanS;
var SnapshotStream = (function (_super) {
    __extends(SnapshotStream, _super);
    function SnapshotStream(behavior, stream) {
        _super.call(this);
        this.behavior = behavior;
        stream.addListener(this);
    }
    SnapshotStream.prototype.push = function (a) {
        this.child.push(behavior_1.at(this.behavior));
    };
    return SnapshotStream;
}(Stream));
function snapshot(b, s) {
    return new SnapshotStream(b, s);
}
exports.snapshot = snapshot;
var SnapshotWithStream = (function (_super) {
    __extends(SnapshotWithStream, _super);
    function SnapshotWithStream(fn, behavior, stream) {
        _super.call(this);
        this.fn = fn;
        this.behavior = behavior;
        stream.child = this;
    }
    SnapshotWithStream.prototype.push = function (a) {
        this.child.push(this.fn(a, behavior_1.at(this.behavior)));
    };
    return SnapshotWithStream;
}(Stream));
function snapshotWith(f, b, s) {
    return new SnapshotWithStream(f, b, s);
}
exports.snapshotWith = snapshotWith;
/** @private */
var SwitchOuter = (function () {
    function SwitchOuter(s) {
        this.s = s;
    }
    ;
    SwitchOuter.prototype.beginPulling = function () {
        throw new Error("not implemented");
    };
    SwitchOuter.prototype.endPulling = function () {
        throw new Error("not implemented");
    };
    SwitchOuter.prototype.push = function (a) {
        this.s.doSwitch(a);
    };
    return SwitchOuter;
}());
var SwitchBehaviorStream = (function (_super) {
    __extends(SwitchBehaviorStream, _super);
    function SwitchBehaviorStream(b) {
        _super.call(this);
        this.b = b;
        this.outerConsumer = new SwitchOuter(this);
        b.addListener(this.outerConsumer);
        var cur = this.currentSource = behavior_1.at(b);
        cur.addListener(this);
    }
    SwitchBehaviorStream.prototype.push = function (a) {
        this.child.push(a);
    };
    SwitchBehaviorStream.prototype.doSwitch = function (newStream) {
        this.currentSource.removeListener(this);
        newStream.addListener(this);
        this.currentSource = newStream;
    };
    return SwitchBehaviorStream;
}(Stream));
function switchStream(b) {
    return new SwitchBehaviorStream(b);
}
exports.switchStream = switchStream;
var ChangesStream = (function (_super) {
    __extends(ChangesStream, _super);
    function ChangesStream(b) {
        _super.call(this);
        this.b = b;
        b.addListener(this);
    }
    ChangesStream.prototype.push = function (a) {
        this.child.push(a);
    };
    ChangesStream.prototype.beginPulling = function () {
        throw new Error("Cannot get changes from pulling behavior");
    };
    ChangesStream.prototype.endPulling = function () {
        throw new Error("Cannot get changes from pulling behavior");
    };
    return ChangesStream;
}(Stream));
function changes(b) {
    return new ChangesStream(b);
}
exports.changes = changes;
var PlaceholderStream = (function (_super) {
    __extends(PlaceholderStream, _super);
    function PlaceholderStream() {
        _super.apply(this, arguments);
    }
    PlaceholderStream.prototype.push = function (a) {
        this.child.push(a);
    };
    PlaceholderStream.prototype.replaceWith = function (s) {
        this.source = s;
        s.addListener(this);
    };
    return PlaceholderStream;
}(Stream));
exports.PlaceholderStream = PlaceholderStream;
function placeholderStream() {
    return new PlaceholderStream();
}
exports.placeholderStream = placeholderStream;
function combineList(ss) {
    // FIXME: More performant implementation with benchmark
    return ss.reduce(function (s1, s2) { return s1.combine(s2); }, empty());
}
exports.combineList = combineList;
/**
 * @returns A stream that never has any occurrences.
 */
function empty() {
    return new SinkStream();
}
exports.empty = empty;
function subscribe(fn, stream) {
    stream.subscribe(fn);
}
exports.subscribe = subscribe;
function combine(a, b) {
    return a.combine(b);
}
exports.combine = combine;
function isStream(obj) {
    return typeof obj === "object" && ("scanS" in obj);
}
exports.isStream = isStream;
//# sourceMappingURL=stream.js.map